# Generated by Django 4.2.8 on 2023-12-22 05:20

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ApiHistory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True, verbose_name='생성일자')),
                ('updated_at', models.DateTimeField(auto_now=True, null=True, verbose_name='수정일자')),
                ('title', models.TextField(blank=True, null=True, verbose_name='제목')),
                ('method', models.TextField(blank=True, null=True, verbose_name='Method')),
                ('url', models.TextField(blank=True, null=True, verbose_name='URL')),
                ('headers', models.JSONField(blank=True, null=True, verbose_name='Header')),
                ('request', models.JSONField(blank=True, null=True, verbose_name='Request')),
                ('response', models.JSONField(blank=True, null=True, verbose_name='Response')),
                ('error', models.TextField(blank=True, null=True, verbose_name='Error')),
            ],
            options={
                'verbose_name': 'history',
                'verbose_name_plural': 'histories',
                'ordering': ['-created_at'],
            },
        ),
        migrations.CreateModel(
            name='NFT',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', model_utils.fields.StatusField(choices=[('PENDING', 'PENDING'), ('APPROVED', 'APPROVED')], default='PENDING', max_length=100, no_check_for_status=True, verbose_name='status')),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, monitor='status', verbose_name='status changed')),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True, verbose_name='생성일자')),
                ('updated_at', models.DateTimeField(auto_now=True, null=True, verbose_name='수정일자')),
                ('token_id', models.IntegerField(null=True, unique=True, verbose_name='token_id')),
                ('request_id', models.UUIDField(null=True, unique=True, verbose_name='request_id')),
                ('name', models.TextField(blank=True, null=True, verbose_name='name')),
                ('address', models.TextField(blank=True, null=True, verbose_name='address')),
                ('nft_type', models.TextField(blank=True, null=True, verbose_name='nft_type')),
                ('attributes', models.TextField(blank=True, null=True, verbose_name='attributes')),
                ('image_url', models.URLField(blank=True, null=True, verbose_name='이미지 URL')),
                ('animation_url', models.URLField(blank=True, null=True, verbose_name='영상 URL')),
                ('external_url', models.URLField(blank=True, null=True, verbose_name='외부 URL')),
                ('extra_url', models.JSONField(blank=True, null=True, verbose_name='추가 URL')),
            ],
            options={
                'verbose_name': 'nft',
                'verbose_name_plural': 'nfts',
                'ordering': ['-created_at'],
            },
        ),
        migrations.CreateModel(
            name='NFTHistory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True, verbose_name='생성일자')),
                ('updated_at', models.DateTimeField(auto_now=True, null=True, verbose_name='수정일자')),
                ('type', models.CharField(choices=[('CREATE', 'CREATE'), ('UPDATE', 'UPDATE'), ('TRANSFER', 'TRANSFER')], max_length=100, verbose_name='유형')),
                ('title', models.TextField(blank=True, null=True, verbose_name='제목')),
                ('nft', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='histories', to='creta.nft')),
            ],
            options={
                'verbose_name': 'nft_history',
                'verbose_name_plural': 'nft_histories',
                'ordering': ['-created_at'],
            },
        ),
    ]



# Python
import base64
import json
from typing import Optional
import logging

import requests
from typing import Any
from urllib.parse import urljoin
import uuid
from datetime import datetime, timezone
import hashlib
import hmac

# Django
from django.conf import settings

# App
from creta.models import ApiHistory, NFT

# Variables
from creta.enum import CollectionType
from creta.utils import *

logger = logging.getLogger(__name__)


# Classes
class BaseGateway:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def get_url(self, path: str) -> str:
        return urljoin(self.base_url, path)

    def get_headers(self, method: str, **kwargs):
        # Set headers
        headers = {
            'x-app-id': settings.CRETA_APP_ID,
            'x-api-key': settings.CRETA_API_KEY,
            'x-timestamp': datetime.now(timezone.utc).isoformat()
        }

        if method == 'POST':
            headers['Content-Type'] = 'application/json'

            request_id = str(uuid.uuid4())
            secret_key = settings.CRETA_SECRET_KEY  # Replace with your actual secret key

            body = json.loads(kwargs['data'])
            body['requestId'] = request_id
            query_string = object_to_query_string(body)

            # Hash the query_string using the secret_key and sha256 algorithm
            signature = base64.b64encode(hmac.new(
                key=secret_key.encode('utf-8'),
                msg=query_string.encode('utf-8'),
                digestmod=hashlib.sha256
            ).digest())
            headers['x-request-id'] = request_id
            headers['x-signature'] = signature

        return headers

    def request(self, title: str, method: str, path: str, **kwargs) -> Any:

        headers = self.get_headers(method=method, **kwargs)
        url = self.get_url(path=path)
        api_history = ApiHistory.objects.create(
            title=title,
            method=method,
            url=url,
            headers=object_to_json(headers),
        )

        # data_str = {key: value.decode('utf-8') if isinstance(value, bytes) else value for key, value in data.items()}
        # json_data = json.dumps(data_str)

        try:
            response = requests.request(method=method, url=url, headers=headers, **kwargs)
            response.raise_for_status()  # 상태 코드가 4xx나 5xx인 경우 예외 발생

        except requests.RequestException as error:
            print('------------------- REQUEST -------------------')
            print('URL: ', response.request.url)
            print('Body: ', response.request.body)
            print('Headers: ', response.request.headers)
            print('-------------------- ERROR --------------------')
            print(error)
            print('--------------------- END ---------------------')
            api_history.error = str(error)
            api_history.save()
            raise error

        # Raise exception for HTTP error status codes such as 400, 404, 500.

        result = response.json()
        api_history.response = response.json()
        api_history.save()

        if result['resultCode'] == 1:
            return result, api_history
        elif result['message']:
            print(result)
            raise Exception(result['message'])
        else:
            print(result)
            raise Exception()


class GatewayV1(BaseGateway):
    def __init__(self):
        super().__init__(base_url=urljoin(settings.CRETA_GATEWAY_HOST, f'/v1/'))

    # (NFT) List By User
    # 유저별 보유 NFT 조회
    # /v1/nfts/assets/{collectionType}/{address}?page=1&limit=30
    # /v1/nfts/assets/200001/0x628fD709BFa7fe68af024852893Ef615104445Ee?page=1&limit=1
    def get_nfts(self, address: str, page=1):
        path = 'nfts/assets/{collectionType}/{address}'.format(collectionType=CollectionType.SUPERCLUB.value, address=address)
        result, api_history = self.request(title="(NFT) List By User", method="GET", path=path, params={'page': page, 'limit': 20})
        return result

    def get_all_nfts(self, address: str):
        all_nfts = []
        page = 1

        while True:
            result = self.get_nfts(address, page)
            # 결과가 빈 배열이면 반복 종료
            if not result["result"]:
                break
            # 현재 페이지의 결과를 누적
            all_nfts.extend(result)
            # 다음 페이지로 이동
            page += 1

        return all_nfts

    def has_nft(self, address: str, token_id: str):
        nfts = self.get_all_nfts(address=address)
        for nft in nfts:
            if nft.get("token_id") == token_id:
                return True
        return False

    # (NFT) Detail
    # 특정 NFT 상세 정보
    # /v1/nfts/detail/{collectionType}/{tokenId}
    # /v1/nfts/detail/200001/10106
    def get_nft(self, token_id: str):
        path = 'nfts/detail/{collectionType}/{tokenId}'.format(collectionType=CollectionType.SUPERCLUB.value, tokenId=token_id)
        result, api_history = self.request(title="(NFT) Detail", method="GET", path=path)
        return result

    # (NFT) Transactions By User
    # 유저별 NFT 거래 내역 조회
    # /v1/tokens/transactions/{address}
    # /v1/nfts/transactions/200001/0x628fD709BFa7fe68af024852893Ef615104445Ee
    def get_nft_transactions_by_user(self, address: str):
        path = 'nfts/transactions/{collectionType}/{address}'.format(collectionType=CollectionType.SUPERCLUB.value, address=address)
        result, api_history = self.request(title="(NFT) Transactions By User", method="GET", path=path)
        return result

    # (NFT) Transactions By NFT
    # 특정 NFT 거래 내역 조회
    # /v1/nfts/history/{collectionType}/{tokenId}
    # /v1/nfts/history/200001/10103
    def get_nft_transactions_by_nft(self, token_id: str):
        path = 'nfts/history/{collectionType}/{tokenId}'.format(collectionType=CollectionType.SUPERCLUB.value, tokenId=token_id)
        result, api_history = self.request(title="(NFT) Transactions By NFT",  method="GET", path=path)
        return result

    # (NFT) Create
    # NFT 민팅
    # /v1/nfts/mint
    def create_nft(self,
            address: str,
            nft_type: int, # Club: 1, User: 2
            name: str,  # 1~63
            image_url: str,  # 10~255
            attributes: Optional[dict] = None,
            animation_url: Optional[str] = None,
            external_url: Optional[str] = None,
            extra_url: Optional[dict] = None):
        path = 'nfts/mint'
        body = {
            "collectionType": CollectionType.SUPERCLUB.value,
            "address": address,
            "nftType": nft_type,
            "name": name,
            "image": image_url,
        }

        optional_keys = ["attributes", "animation_url", "external_url", "extra_url"]
        for key in optional_keys:
            value = locals()[key]
            if value is not None:
                body[key] = value

        result, api_history = self.request(title="(NFT) Create", method="POST", path=path, data=json.dumps(body))
        request_id = api_history.headers['x-request-id']
        result['request_id'] = request_id

        instance = NFT.objects.create(
            name=name,
            nft_type=nft_type,
            attributes=attributes,
            request_id=request_id,
            image_url=image_url,
            animation_url=animation_url,
            external_url=external_url,
            extra_url=extra_url
        )

        return result, instance

    # (Nft) Status
    # NFT 민팅 현황 조회
    # /v1/nfts/mint/{collectionType}/{requestId}
    # /v1/nfts/mint/200001/3c52f871-1a01-47f3-8a9e-045d2767aab9
    def check_nft(self, request_id):
        path = 'nfts/mint/{collectionType}/{requestId}'.format(collectionType=200001, requestId=request_id)
        result, api_history = self.request(title="(NFT) Status", method="GET", path=path)

        instance = NFT.objects.get(request_id=request_id)

        if result and result['result'] and result['result']['tokenId']:
            instance.token_id = result['result']['tokenId']
            instance.status = 'APPROVED'
            instance.save()

        return result, instance

    # (NFT) Update
    # NFT 데이터 수정
    # /v1/nfts/update
    def update_nft(self, token_id: str, nft_type: str, name: str, attributes):
        path = 'nfts/update'
        body = {
            "collectionType": CollectionType.SUPERCLUB.value,
            "tokenId": token_id,
            "nftType": nft_type,
            "name": name,
            "attributes": attributes
        }

        result, api_history = self.request(title="(NFT) Update", method="POST", path=path, data=json.dumps(body))

        instance = NFT.objects.get(token_id=int(token_id))

        instance.name = name
        instance.nft_type = nft_type
        instance.attributes = attributes
        # instance.image_url = image_url
        # instance.animation_url = animation_url
        # instance.external_url = external_url
        # instance.extra_url = extra_url
        instance.save()

        return result, instance

    # (NFT) Transfer
    # 다른 유저에게 NFT 전송
    # /v1/nfts/transfer
    def transfer_nft(self, address_from: str, address_to: str, token_id: str):
        path = 'nfts/transfer'
        body = {
            "collectionType": CollectionType.SUPERCLUB.value,
            "from": address_from,
            "to": address_to,
            "tokenId": token_id
        }
        result, api_history = self.request(title="(NFT) transfer", method="POST", path=path, data=json.dumps(body))
        return result


# Instances
creta_sdk = GatewayV1()



# vim: set fileencoding=utf-8 :
from django.contrib import admin, messages

# App
from creta import models


# Main Section
class BaseModelAdmin(admin.ModelAdmin):
    actions = []


class NFTAdmin(BaseModelAdmin):
    list_filter = ()


class NFTHistoryAdmin(BaseModelAdmin):
    list_filter = ()


class ApiHistoryAdmin(BaseModelAdmin):
    list_filter = ()


def _register(model, admin_class):
    admin.site.register(model, admin_class)


_register(models.NFT, NFTAdmin)
_register(models.NFTHistory, NFTHistoryAdmin)
_register(models.ApiHistory, ApiHistoryAdmin)



# Django Rest Framework
from rest_framework import serializers

# App
from creta.models import ApiHistory, NFTHistory, NFT


# Classes
class NFTSerializer(serializers.ModelSerializer):

    class Meta:
        model = NFT
        fields = ['id', 'token_id', 'request_id',
                  'name', 'address', 'nft_type', 'attributes',
                  'image_url', 'animation_url', 'external_url', 'extra_url',
                  'created_at', 'updated_at']


class NftStatusSerializer(serializers.ModelSerializer):

    class Meta:
        model = NFT
        fields = ('request_id',)


class NFTHistorySerializer(serializers.ModelSerializer):

    class Meta:
        model = NFTHistory
        fields = ['id', 'nft', 'type', 'title', 'created_at', 'updated_at']


class ApiHistorySerializer(serializers.ModelSerializer):

    class Meta:
        model = ApiHistory
        fields = ['id', 'title', 'method', 'url', 'headers', 'request', 'response', 'error', 'created_at', 'updated_at']


import json
from urllib.parse import urlencode, quote


def object_to_query_string(params):
    encode_params = {}
    params = dict(sorted(params.items()))

    def _encode_params(params, p_key=None):
        if isinstance(params, dict):
            for key, value in params.items():
                encode_key = f"{p_key}[{key}]" if p_key else key
                _encode_params(value, encode_key)
        elif isinstance(params, (list, tuple)):
            for offset, value in enumerate(params):
                encode_key = f"{p_key}[{offset}]"
                _encode_params(value, encode_key)
        else:
            encode_params[p_key] = params

    if isinstance(params, dict):
        for key, value in params.items():
            _encode_params(value, key)

    return urlencode(encode_params, quote_via=quote)


def convert_bytes_to_str(obj):
    if isinstance(obj, bytes):
        return obj.decode('utf-8')
    elif isinstance(obj, list):
        return [convert_bytes_to_str(item) for item in obj]
    elif isinstance(obj, dict):
        return {key: convert_bytes_to_str(value) for key, value in obj.items()}
    else:
        return obj


def object_to_json(obj):
    return convert_bytes_to_str(obj)



# Django
from django.db import models

# 3rd Party
from model_utils.models import StatusModel
from model_utils import Choices


# Class Section
class BaseModel(models.Model):
    # Dates
    created_at = models.DateTimeField('생성일자', auto_now_add=True, null=True)
    updated_at = models.DateTimeField('수정일자', auto_now=True, null=True)

    class Meta:
        abstract = True


class NFT(BaseModel, StatusModel):
    STATUS = Choices(
        ('PENDING', 'PENDING'),
        ('APPROVED', 'APPROVED'),
    )

    # Unique
    token_id = models.IntegerField('token_id', unique=True, null=True)
    request_id = models.UUIDField('request_id', unique=True, null=True)

    # Basic
    name = models.TextField('name', blank=True, null=True)
    address = models.TextField('address', blank=True, null=True)
    nft_type = models.TextField('nft_type', blank=True, null=True)
    attributes = models.TextField('attributes', blank=True, null=True)

    # URL
    image_url = models.URLField('이미지 URL', null=True, blank=True)
    animation_url = models.URLField('영상 URL', null=True, blank=True)
    external_url = models.URLField('외부 URL', null=True, blank=True)
    extra_url = models.JSONField('추가 URL', blank=True, null=True)

    class Meta:
        verbose_name = 'nft'
        verbose_name_plural = 'nfts'
        ordering = ['-created_at']

    def clean(self):
        if self.request_id == "":
            self.request_id = None

    def save(self, *args, **kwargs):
        if self.token_id:
            self.status = 'APPROVED'
        else:
            self.status = 'PENDING'

        return super(NFT, self).save(*args, **kwargs)


class NFTHistory(BaseModel):
    nft = models.ForeignKey('creta.NFT', on_delete=models.CASCADE, related_name='histories')
    type = models.CharField('유형', max_length=100, choices=[('CREATE', 'CREATE'), ('UPDATE', 'UPDATE'), ('TRANSFER', 'TRANSFER'), ])
    title = models.TextField('제목', blank=True, null=True)

    class Meta:
        verbose_name = 'nft_history'
        verbose_name_plural = 'nft_histories'
        ordering = ['-created_at']


class ApiHistory(BaseModel):
    title = models.TextField('제목', blank=True, null=True)
    method = models.TextField('Method', blank=True, null=True)
    url = models.TextField('URL', blank=True, null=True)
    headers = models.JSONField('Header', blank=True, null=True)
    request = models.JSONField('Request', blank=True, null=True)
    response = models.JSONField('Response', blank=True, null=True)
    error = models.TextField('Error', blank=True, null=True)

    class Meta:
        verbose_name = 'history'
        verbose_name_plural = 'histories'
        ordering = ['-created_at']





from django.urls import path
from rest_framework.routers import DefaultRouter

# App
from creta.views import ApiHistoryViewSet


# Variables
router = DefaultRouter()
router.register(r'histories', ApiHistoryViewSet)
urlpatterns = router.urls

urlpatterns += []



from django.test import TestCase

# App
from creta.gateway import creta_sdk


# Main
class GatewayV1Test(TestCase):
    def test_get_nfts(self):
        result = creta_sdk.get_nfts(address="0x628fD709BFa7fe68af024852893Ef615104445Ee", page=1)
        self.assertEqual(result['resultCode'], 1)

    def test_get_nft(self):
        result = creta_sdk.get_nft(token_id="10106")
        self.assertEqual(result['resultCode'], 1)

    def test_get_nft_transactions_by_user(self):
        result = creta_sdk.get_nft_transactions_by_user(address="0x628fD709BFa7fe68af024852893Ef615104445Ee")
        self.assertEqual(result['resultCode'], 1)

    def test_get_nft_transactions_by_nft(self):
        result = creta_sdk.get_nft_transactions_by_nft(token_id="10103")
        self.assertEqual(result['resultCode'], 1)

    def test_create_nft(self):
        result, instance = creta_sdk.create_nft(
            address="0x628fD709BFa7fe68af024852893Ef615104445Ee",
            nft_type=1,
            name="NFT name",
            attributes=[
                {
                    "trait_type": "Level",
                    "value": 1
                }
            ],
            image_url="https://cf.dev.superclubs.io/media/club/advertisement/8f4ee8fd-8c79-4004-aad7-88c3edbdf4ef/20230907/20230907T040903.jpg",
        )
        print('test_create_nft: ', result)

    def test_check_nft(self):
        result, instance = creta_sdk.check_nft(request_id="3c52f871-1a01-47f3-8a9e-045d2767aab9")
        self.assertEqual(result['resultCode'], 1)

    def test_update_nft(self):
        result, instance = creta_sdk.update_nft(
            token_id="10106",
            nft_type=3,
            name="Club 20",
            attributes=[{"trait_type": "Level", "value": 20}])
        self.assertEqual(result['resultCode'], 1)

    def test_transfer_nft(self):
        result = creta_sdk.transfer_nft(
            address_from="0x628fD709BFa7fe68af024852893Ef615104445Ee",
            address_to="0x7f369511e314f84e16b08413547a04e3a9552508",
            token_id="10103"
        )
        self.assertEqual(result['resultCode'], 1)


# Django Rest Framework
from rest_framework import mixins, filters
from rest_framework.viewsets import GenericViewSet

# Third Party
from django_filters.rest_framework import DjangoFilterBackend

# App
from creta import pagination
from creta.models import ApiHistory
from creta.serializers import ApiHistorySerializer

# Variables
name_search_fields = ['name', 'name_en', 'name_ja', 'name_ko']


# Classes
class BaseGenericViewSet(GenericViewSet):
    search_fields = name_search_fields
    filter_backends = [filters.OrderingFilter, filters.SearchFilter, DjangoFilterBackend]
    pagination_class = pagination.DefaultPagination


class ApiHistoryViewSet(
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    BaseGenericViewSet
):
    queryset = ApiHistory.objects.all()
    serializer_class = ApiHistorySerializer

    search_fields = ('title',)
    ordering_fields = ('id', 'created',)
    filterset_fields = []

